
#ifndef _FreeOTFE_H
#define _FreeOTFE_H   1




// The following includes are required in order to implement Cancel-Safe
// IRP Queues
#include <ntddk.h>
//#include <wdm.h>
#include <csq.h>

#include "FreeOTFECypherAPI.h"
#include "FreeOTFEHashAPI.h"


// =========================================================================
// Const definitions


// These values are in DECIMAL
#define DRIVER_VERSION_MAJOR     00
#define DRIVER_VERSION_MINOR     59
#define DRIVER_VERSION_BUILD   0000

#define DRIVER_VERSION  (                                      \
                         (DRIVER_VERSION_MAJOR * 0x01000000) + \
                         (DRIVER_VERSION_MINOR * 0x00010000) + \
                         (DRIVER_VERSION_BUILD * 0x00000001)   \
                        )

#define MAX_DEVICES 26


#define HDD_BYTES_PER_SECTOR 512
#define CD_BYTES_PER_SECTOR  2048
// These figures are used; no real reason for these *exact* numbers, except
// that they're the same defaults as used by Windows NT for SCSI devices
// when it can't get the real values from the SCSI BIOS.
// See MS Knowledgebase article Q258281 for further details
// Tracks per cylinder = the number of heads
#define TRACKS_PER_CYLINDER 64
#define SECTORS_PER_TRACK 32

// Encrypt in 512 byte "sectors"
#define ENCRYPTION_BLOCK_SIZE 512

// Maximum MAC length
// This should be set to the MAXIMUM MAC length that the any MAC may generate.
// Note: This value is in *bits*
#define FREEOTFE_MAX_MAC_LENGTH 1024


// Definition used for testing purposes; dummy GUID with 0x0F as it's last byte
DEFINE_GUID(TEST_GUID_0x0F,  0x00000000L, 0x0000, 0x0000, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0F);


#define CD_FRAMES_PER_SECOND 75
#define CD_FRAMES_PER_MINUTE (CD_FRAMES_PER_SECOND * 60)


// =========================================================================
// Type definitions

   

typedef struct _DEVICE_EXTENSION {

    // Allow differentiation between the main device and disk devices
    BOOLEAN IsMainDevice;
    
    // Device's name
    UNICODE_STRING zzDeviceName;

    // ------------------------------------------------------
    // Main device items ONLY
    
    // Symbolic link name
    UNICODE_STRING zzSymbolicLinkName;
    
        
    // ------------------------------------------------------
    // Disk device items ONLY
    
    // Flag if a volume is mounted or not
    BOOLEAN Mounted;

    // Flag if a volume is being dismounted or not
    // i.e. This flag determines if any more IRPs should be queued for
    // the thread or not. If set, no further IRPs will be accepted
    BOOLEAN DismountPending;
    
    // Filename of any mounted volume
    UNICODE_STRING zzFilename;

    // Handle to the file
    HANDLE FileHandle;
	PFILE_OBJECT FileObject;

    // Flag if file attributes have been stored (e.g. FALSE for partitions, etc)
    BOOLEAN FileAttributesStored;
    // If stored, the file timestamps/attributes when opened
    FILE_BASIC_INFORMATION FileAttributes;


    // The "sector size" in which blocks should actually be read/written to the volume
    // file/partition
    DWORD FileSectorSize;
    // DataOffset % DiskGeometry->SectorSize
    DWORD DataOffsetModVirtualSectorSize;

    
    // Start of encrypted data within the file
    // Note: We don't need to store the end offset, as this can be determined by using the
    //       DiskGrometry/DiskSize members
    LARGE_INTEGER DataOffset;

    // Simulated disk geometry
    DISK_GEOMETRY DiskGeometry;
    LARGE_INTEGER DiskSize;
    LARGE_INTEGER PartitionSize;

    // Count of hidden sectors (if any)
    DWORD HiddenSectors;

    // Encryption block size
    // This *should* always be set to the sector size of the emulated device, but isn't always
    // (e.g. Linux ISO images have an emulated 2048 byte sector size, but encrypt in 512 byte
    // blocks)
    DWORD EncryptionBlockSize;

    // Readonly flag
    BOOLEAN ReadOnly;

    // Mount source (e.g. the volume file is a partition or file)
    MOUNT_SOURCE MountSource;

    // Storage media type (i.e. the type of device emulated)
    STORAGE_MEDIA_TYPE StorageMediaType;

    // Prevent media removal; for removable disks only
    BOOLEAN PreventMediaRemoval;
        

    // -----
	// IV Hash device ID
    UNICODE_STRING IVHashDeviceName;
    GUID IVHashGUID;

	// IV Hash device handle
	PFILE_OBJECT IVHashFileObject;
	PDEVICE_OBJECT IVHashDeviceObject;
	
	// IV Hash device internal details
    HASH IVHashDetails;
    PDataHashFn FnIVHash;

    // -----
	// IV cypher device ID
    UNICODE_STRING IVCypherDeviceName;
    GUID IVCypherGUID;

	// IV cypher device handle
	PFILE_OBJECT IVCypherFileObject;
	PDEVICE_OBJECT IVCypherDeviceObject;
	
	// IV cypher device internal details
	CYPHER IVCypherDetails;
	PDataEncryptFn FnIVEncrypt;
    PDataDecryptFn FnIVDecrypt;

    // -----
    // Main cypher device ID
    UNICODE_STRING MainCypherDeviceName;
    GUID MainCypherGUID;

    // Main cypher device handle
    PFILE_OBJECT MainCypherFileObject;
    PDEVICE_OBJECT MainCypherDeviceObject;
    
	// Main cypher device internal details
    CYPHER MainCypherDetails;
	PDataEncryptFn FnMainEncrypt;
    PDataDecryptFn FnMainDecrypt;

    // -----
    // Key to be used for encryption/decryption
    ULONG MasterKeyLength;  // This value is in *bits*
    unsigned char *MasterKey;
    // MasterKeyASCII is the ASCII representation of MasterKey
    // This is stored as all AES candidates use this format for supplying
    // their keys (blame NIST for that...)
    unsigned char *MasterKeyASCII;  // Hex ASCIIZ nibbles

    // Key to be used for ESSIV generation (*if* *required*)
    ULONG ESSIVKeyLength;  // This value is in *bits*
    unsigned char *ESSIVKey;
    // ESSIVKeyASCII is the ASCII representation of ESSIVKey
    // This is stored as all AES candidates use this format for supplying
    // their keys (blame NIST for that...)
    unsigned char *ESSIVKeyASCII;  // Hex ASCIIZ nibbles

    // Volume IV to be used to encrypt/decrypt each sector
    ULONG VolumeIVLength;  // This value is in *bits*
    unsigned char *VolumeIV;

    // -----
    // Various flags
    unsigned int VolumeFlags;

    // Sector IV generation method
    SECTOR_IV_GEN_METHOD SectorIVGenMethod;

    // Metadata
    // This is used to store arbitary user-mode data when the volume is mounted.
    // This data will be returned whenever the status is the disk is returned
    ULONG MetaDataLength;  // This value is in *bytes*
    unsigned char *MetaData;

    // ------------------------------------------------------
    // Thread/IRP queue related items follow    
    
    // Flag to signal that the thread should terminate
    BOOLEAN TerminateThread;
    
    // The device's IRP processing thread
    PETHREAD ThreadObject;
    
    // Irps waiting to be processed are queued here
    LIST_ENTRY   PendingIRPQueue;

    //  SpinLock to protect access to the queue
    KSPIN_LOCK IRPQueueLock;
    
    IO_CSQ CancelSafeQueue;   
    KSEMAPHORE IRPQueueSemaphore;


    // ------------------------------------------------------
 
  
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;


// =========================================================================
// Function headers

// Driver entry point
NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
);
    
    
// Driver unload
VOID
DriverUnload(
    IN PDRIVER_OBJECT DriverObject
);
    
    
// Handle create IRPs
NTSTATUS
FreeOTFE_MF_DispatchCreate(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


// Handle close IRPs
NTSTATUS
FreeOTFE_MF_DispatchClose(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


// Handle device control IRPs
NTSTATUS
FreeOTFE_MF_DispatchDeviceControl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


// Handle read IRPs
NTSTATUS
FreeOTFE_MF_DispatchRead(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


// Handle write IRPs
NTSTATUS
FreeOTFE_MF_DispatchWrite(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


// Handle flush buffers IRPs
NTSTATUS
FreeOTFE_MF_DispatchFlushBuffers(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);

    
// Handle power IRPs
NTSTATUS
FreeOTFE_MF_DispatchPower(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


// Handle system control IRPs
NTSTATUS
FreeOTFE_MF_DispatchSystemControl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);

NTSTATUS
CreateMainDevice (
    IN PDRIVER_OBJECT   DriverObject,
    OUT PDEVICE_OBJECT  *DeviceObject
);


NTSTATUS
CreateDiskDevice (
    IN PDRIVER_OBJECT   DriverObject,
    IN ULONG DeviceType,
    OUT PDEVICE_OBJECT  *DeviceObject
);
    

PDEVICE_OBJECT
DestroyDevice(
    IN PDEVICE_OBJECT devObj    
);

    
VOID
FreeOTFEThread(
    IN PVOID Context
);
    
    
    
    

VOID CSQInsertIrp (
    IN PIO_CSQ   Csq,
    IN PIRP              Irp
);

VOID CSQRemoveIrp(
    IN  PIO_CSQ Csq,
    IN  PIRP    Irp
);

PIRP CSQPeekNextIrp(
    IN  PIO_CSQ Csq,
    IN  PIRP    Irp,
    IN  PVOID   PeekContext
);

VOID CSQAcquireLock(
    IN  PIO_CSQ Csq,
    OUT PKIRQL  Irql
);

VOID CSQReleaseLock(
    IN PIO_CSQ Csq,
    IN KIRQL   Irql
);

VOID CSQCompleteCanceledIrp(
    IN  PIO_CSQ             pCsq,
    IN  PIRP                Irp
);

// Cancel all IRPs currently on the queue
void
CancelAllQueuedIRPs(
    IN PDEVICE_OBJECT DeviceObject
);


// Queue the specified IRP for the given DeviceObject
NTSTATUS
QueueIRPToThread(
    IN  PDEVICE_OBJECT DeviceObject,    
    IN  PIRP           Irp
);


NTSTATUS
GetNamedCharDevice(
    IN PDRIVER_OBJECT DriverObject,
    IN PCHAR TgtDeviceName, 
    OUT PDEVICE_OBJECT* DeviceObject
);
    
    
NTSTATUS
GetNamedUnicodeDevice(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING TgtDeviceName, 
    OUT PDEVICE_OBJECT* DeviceObject
);


// Read/write data to the specified device
NTSTATUS
ThreadReadWriteData(
    IN PDEVICE_OBJECT DeviceObject,
    IN OUT PIRP Irp
);


// Mount the specified device
NTSTATUS
ThreadMount(
    IN PDEVICE_OBJECT DeviceObject,
    IN OUT PIRP Irp
);


// Dismount the specified device
NTSTATUS
ThreadDismount(
    IN PDEVICE_OBJECT DeviceObject,
    IN OUT PIRP Irp
);


NTSTATUS
IOCTL_FreeOTFEIOCTL_Version(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_FreeOTFEIOCTL_Create(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);
    
    
NTSTATUS
IOCTL_FreeOTFEIOCTL_Destroy(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


// Write raw data to volume
NTSTATUS
IOCTL_FreeOTFEIOCTL_SetRaw(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);

    
// Read raw data from volume
NTSTATUS
IOCTL_FreeOTFEIOCTL_GetRaw(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);

// Derive a key
// This is implemented in the kernel mode driver as implementing it in userspace (e.g. the
// Delphi frontend) takes *far* too long to process the data due to the repeated calls to hash
// functions, etc
NTSTATUS
IOCTL_FreeOTFEIOCTL_DeriveKey(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);

    
// Generate a MAC
// Implmented in the kernel mode driver because it's required by the key derivation process
NTSTATUS
IOCTL_FreeOTFEIOCTL_GenerateMAC(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_FreeOTFEIOCTL_Mount(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PDEVICE_OBJECT *QueueDeviceObject
);
    
    
NTSTATUS
IOCTL_FreeOTFEIOCTL_Dismount(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PDEVICE_OBJECT *QueueDeviceObject
);


NTSTATUS
IOCTL_FreeOTFEIOCTL_GetDiskDeviceCount(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_FreeOTFEIOCTL_GetDiskDeviceList(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_FreeOTFEIOCTL_GetDiskDeviceStatus(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);
      

NTSTATUS
IOCTL_FreeOTFEIOCTL_GetDiskDeviceMetaData(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);

NTSTATUS
IOCTL_Std_DiskFormatTracks(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_DiskIsWritable(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_DiskGetDriveGeometry(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_DiskGetDriveLayout(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_DiskGetDriveLayoutEx(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_DiskGetLengthInfo(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_DiskGetPartitionInfo(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_DiskGetPartitionInfoEx(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_DiskSetPartitionInfo(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_DiskSetPartitionInfoEx(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_DiskVerify(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_CheckVerify(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_MediaRemoval(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
IOCTL_Std_StorageEjectMedia(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PDEVICE_OBJECT *QueueDeviceObject
);


NTSTATUS
IOCTL_Std_CDROMReadTOC(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
);


NTSTATUS
GetDeviceDetailsHash(
    IN      char* deviceName,
    IN      GUID* supportGUID,

    OUT     UNICODE_STRING* HashDeviceName,
	OUT     PFILE_OBJECT* HashFileObject,
	OUT     PDEVICE_OBJECT* HashDeviceObject,
    OUT     HASH* HashDetails,
    OUT     PDataHashFn* FnHash
);

// Overwrite, free off and release everything that may have been
// created/obtained by a prior call to GetDeviceDetailsHash(...)
void
FreeDeviceDetailsHash(
    OUT     UNICODE_STRING* HashDeviceName,
	OUT     PFILE_OBJECT* HashFileObject,
	OUT     PDEVICE_OBJECT* HashDeviceObject,
    OUT     HASH* HashDetails,
    OUT     PDataHashFn* FnHash
);


NTSTATUS
GetDeviceDetailsCypher(
    IN      char* deviceName,
    IN      GUID* supportGUID,

    OUT     UNICODE_STRING* CypherDeviceName,
    OUT     PFILE_OBJECT* CypherFileObject,
    OUT     PDEVICE_OBJECT* CypherDeviceObject,
    OUT     CYPHER* CypherDetails,
	OUT     PDataEncryptFn* FnEncrypt,
    OUT     PDataDecryptFn* FnDecrypt
);

// Overwrite, free off and release everything that may have been
// created/obtained by a prior call to GetDeviceDetailsCypher(...)
void
FreeDeviceDetailsCypher(
    OUT     UNICODE_STRING* CypherDeviceName,
    OUT     PFILE_OBJECT* CypherFileObject,
    OUT     PDEVICE_OBJECT* CypherDeviceObject,
    OUT     CYPHER* CypherDetails,
	OUT     PDataEncryptFn* FnEncrypt,
    OUT     PDataDecryptFn* FnDecrypt
);


NTSTATUS
SynchCompletionRoutine(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PKEVENT Event
);


// Call the relevant hash function
NTSTATUS
IVDataHash(
    IN      PDEVICE_OBJECT DeviceObject,
    IN      unsigned int dataLength,  // In bits
    IN      PVOID dataBuffer,
    IN OUT  unsigned int *hashLength,  // In bits
    IN      PVOID hashBuffer    
);


NTSTATUS
GenerateBlockIV(
    IN   PDEVICE_OBJECT DeviceObject,
    IN   LARGE_INTEGER fileOffset,  // The offset within the *volume file* where
                                    // the data was obtained from
    IN   unsigned int blockIVLength,  // In bits
    OUT  unsigned char* blockIV
);


// Generate block IV: Sector ID
NTSTATUS
GenerateBlockIV_SectorID(
    IN      LARGE_INTEGER blockID,
    IN      unsigned int blockIDBits,  // The number of bits in blockID to be used
                                       // Set to either 32 or 64 bits
    IN      unsigned int blockIVLength,  // The size of the required IV (and length of
	                                     // "blockIV"), in bits
    OUT     unsigned char* blockIV  // Pointer to where the IV should be written
                                    // Note: Caller must pre-initialise to 0x00s
);

// Generate block IV: Hashed sector ID
NTSTATUS
GenerateBlockIV_HashedSectorID(
    IN      PDEVICE_OBJECT DeviceObject,
    IN      LARGE_INTEGER blockID,
    IN      unsigned int blockIDBits,  // The number of bits in blockID to be used
                                       // Set to either 32 or 64 bits
    IN      unsigned int blockIVLength,  // The size of the required IV (and length of
	                                     // "blockIV"), in bits
    OUT     unsigned char* blockIV  // Pointer to where the IV should be written
                                    // Note: Data here will be pre-initialised to 0x00s
);

// Generate block IV: ESS_IV
NTSTATUS
GenerateBlockIV_ESSIV(
    IN      PDEVICE_OBJECT DeviceObject,
    IN      LARGE_INTEGER blockID,
    IN      unsigned int blockIDBits,  // The number of bits in blockID to be used
                                       // Set to either 32 or 64 bits
    IN      unsigned int blockIVLength,  // The size of the required IV (and length of
	                                     // "blockIV"), in bits
    OUT     unsigned char* blockIV  // Pointer to where the IV should be written
                                    // Note: Data here will be pre-initialised to 0x00s
);


// Call the relevant encryption function
NTSTATUS
BlockEncrypt(
    IN PDEVICE_OBJECT DeviceObject,
    IN LARGE_INTEGER fileOffset,  // The offset within the *volume file* where the
                                  // data was obtained from
    IN PBYTE plaintextBuffer,
    IN PBYTE cyphertextBuffer
);
  
    
// Call the relevant decryption function
NTSTATUS
BlockDecrypt(
    IN PDEVICE_OBJECT DeviceObject,
    IN LARGE_INTEGER fileOffset,  // The offset within the *volume file* where the
                                  // data was obtained from
    IN PBYTE cyphertextBuffer,
    IN PBYTE plaintextBuffer
);


// Overwrite sensitive parts of the specified device's extension
NTSTATUS
OverwriteDeviceSensitive(
    IN  PDEVICE_OBJECT devObj
);


    
// Write/Read raw data to/from volume
// DoWrite - set to TRUE to write, FALSE to get
NTSTATUS
SetGetRawChar(
    IN     BOOLEAN DoWrite,
    IN     PCHAR Filename,
    IN     LARGE_INTEGER Offset,
    IN     ULONG DataLength,
    IN OUT BYTE *Data
);
    
// Write/Read raw data to/from volume
// DoWrite - set to TRUE to write, FALSE to get
NTSTATUS
SetGetRawUnicode(
    IN     BOOLEAN DoWrite,
    IN     PUNICODE_STRING Filename,
    IN     LARGE_INTEGER Offset,
    IN     ULONG DataLength,
    IN OUT BYTE *Data
);
    
    
// Determine the max size of a file
NTSTATUS
GetMaxSizeFile(
    IN  HANDLE FileHandle,
    OUT PLARGE_INTEGER MaxSize
);

// Determine the max size of a partition
// DeviceName - The device name of the partition
//              e.g. \Device\Harddisk1\Partition1
NTSTATUS
GetMaxSizePartition(
    IN  PUNICODE_STRING DeviceName,
    OUT PLARGE_INTEGER MaxSize
);


NTSTATUS
GenerateMAC(
    IN      MAC_ALGORITHM MACAlgorithm,
    IN      CHAR HashDeviceName[FREEOTFE_MAX_FILENAME_LENGTH],
    IN      GUID HashGUID,
    IN      CHAR CypherDeviceName[FREEOTFE_MAX_FILENAME_LENGTH],
    IN      GUID CypherGUID,
    IN      int LengthWanted,  // In bits
    IN      unsigned int KeyLength,  // In bits
    IN      unsigned char* Key,
    IN      unsigned int DataLength,  // In bits
    IN      unsigned char* Data,

    IN OUT  unsigned int* MACLength,  // In bits
    OUT     unsigned char* MAC
);


NTSTATUS
DeriveKey(
    IN      KDF_ALGORITHM KDFAlgorithm,
    IN      CHAR HashDeviceName[FREEOTFE_MAX_FILENAME_LENGTH],
    IN      GUID HashGUID,
    IN      CHAR CypherDeviceName[FREEOTFE_MAX_FILENAME_LENGTH],
    IN      GUID CypherGUID,
    IN      int Iterations,
    IN      int LengthWanted,
    IN      unsigned int PasswordLength,  // In bits
    IN      unsigned char* Password,
    IN      unsigned int SaltLength,  // In bits
    IN      unsigned char* Salt,

    IN OUT  unsigned int* DerivedKeyLength,  // In bits
    OUT     unsigned char* DerivedKey
);

// Perform an actual read from the encrypted volume
// This can be supplied with an arbitary Offset and BufferLength
NTSTATUS
ActualReadFile(
    IN      PDEVICE_OBJECT DeviceObject,
    IN      PLARGE_INTEGER Offset,
    IN      ULONG BufferLength,
    IN      PBYTE Buffer
);

// Perform an actual write to the encrypted volume
// This can be supplied with an arbitary Offset and BufferLength
NTSTATUS
ActualWriteFile(
    IN      PDEVICE_OBJECT DeviceObject,
    IN      PLARGE_INTEGER Offset,
    IN      ULONG BufferLength,
    OUT     PBYTE Buffer
);

// Populate PARTITION_INFORMATION/PARTITION_INFORMATION_EX structures
NTSTATUS
PopulatePartitionInfo(
    IN PDEVICE_OBJECT DeviceObject,
    OUT PPARTITION_INFORMATION PartitionInfo,
    OUT PPARTITION_INFORMATION_EX PartitionInfoEx
);

// Identify if emulating a CD or DVD device
BOOLEAN
EmulatingCDOrDVD(
    IN PDEVICE_OBJECT DeviceObject
);

// This function is for testing new functionality ONLY
NTSTATUS TestCode();


// =========================================================================
// =========================================================================
// =========================================================================

#endif


